import io
from unittest import mock, TestCase

from pcs_test.tools.custom_mock import (
    MockCurl,
    MockCurlMulti,
)

from pcs import settings
from pcs.common import (
    host,
    pcs_pycurl as pycurl,
)
from pcs.common.host import Destination
import pcs.common.node_communicator as lib

PORT = settings.pcsd_default_port


class NodeTargetFactory(TestCase):
    def setUp(self):
        self.known_name = "node"
        self.unknown_name = "none"
        self.known_host = host.PcsKnownHost(
            self.known_name, "token", [host.Destination("addr", "port")]
        )
        self.factory = lib.NodeTargetFactory({self.known_name: self.known_host})

    def assert_equal_known_host_target(self, known_host, target):
        self.assertEqual(known_host.name, target.label)
        self.assertEqual(known_host.token, target.token)
        self.assertEqual(known_host.dest_list, target.dest_list)

    def test_get_target_success(self):
        self.assert_equal_known_host_target(
            self.known_host, self.factory.get_target(self.known_name)
        )

    def test_get_target_not_found(self):
        # pylint: disable=invalid-name
        with self.assertRaises(lib.HostNotFound) as cm:
            self.factory.get_target(self.unknown_name)
        self.assertEqual(self.unknown_name, cm.exception.name)

    def test_from_hostname_known(self):
        self.assert_equal_known_host_target(
            self.known_host,
            self.factory.get_target_from_hostname(self.known_name)
        )

    def test_from_hostname_unknown(self):
        target = self.factory.get_target_from_hostname(self.unknown_name)
        self.assertEqual(self.unknown_name, target.label)
        self.assertEqual(None, target.token)
        self.assertEqual(
            [host.Destination(self.unknown_name, PORT)], target.dest_list
        )


class RequestDataUrlEncodeTest(TestCase):
    def test_no_data(self):
        action = "action"
        data = lib.RequestData(action)
        self.assertEqual(action, data.action)
        self.assertEqual(0, len(data.structured_data))
        self.assertEqual("", data.data)

    def test_with_data(self):
        action = "action"
        orig_data = [
            ("key1", "value1"),
            ("spacial characters", "+-+/%&?'\";[]()*^$#@!~`{:}<>")
        ]
        data = lib.RequestData(action, orig_data)
        self.assertEqual(action, data.action)
        self.assertEqual(orig_data, data.structured_data)
        expected_raw_data_variants = (
            "key1=value1&spacial+characters=%2B-%2B%2F%25%26%3F%27%22%3B%5B"
                "%5D%28%29%2A%5E%24%23%40%21%7E%60%7B%3A%7D%3C%3E"
            ,
            "key1=value1&spacial+characters=%2B-%2B%2F%25%26%3F%27%22%3B%5B"
                "%5D%28%29%2A%5E%24%23%40%21~%60%7B%3A%7D%3C%3E"
            ,
        )
        self.assertTrue(data.data in expected_raw_data_variants)


def _addr_list_to_dest(addr_list, port=None):
    return [Destination(addr, port) for addr in addr_list]


class RequestTargetConstructorTest(TestCase):
    def test_no_adresses(self):
        label = "label"
        target = lib.RequestTarget(label)
        self.assertEqual(label, target.label)
        self.assertEqual(
            _addr_list_to_dest([label], port=PORT), target.dest_list
        )

    def test_with_adresses(self):
        label = "label"
        address_list = ["a1", "a2"]
        original_list = list(address_list)
        target = lib.RequestTarget(
            label, dest_list=_addr_list_to_dest(address_list)
        )
        address_list.append("a3")
        self.assertEqual(label, target.label)
        self.assertIsNot(_addr_list_to_dest(address_list), target.dest_list)
        self.assertEqual(_addr_list_to_dest(original_list), target.dest_list)


class RequestTargetFromKnownHost(TestCase):
    def test_success(self):
        known_host = host.PcsKnownHost(
            "name",
            "token",
            [
                host.Destination("addr1", "addr2"),
                host.Destination("addr2", "port2"),
            ]
        )
        target = lib.RequestTarget.from_known_host(known_host)
        self.assertEqual(known_host.name, target.label)
        self.assertEqual(known_host.token, target.token)
        self.assertEqual(known_host.dest_list, target.dest_list)



class RequestUrlTest(TestCase):
    action = "action"

    def _get_request(self, target):
        return lib.Request(target, lib.RequestData(self.action))

    def assert_url(self, actual_url, hostname, action, port=None):
        if port is None:
            port = settings.pcsd_default_port
        self.assertEqual(
            f"https://{hostname}:{port}/{action}",
            actual_url
        )

    def test_url_basic(self):
        hostname = "host"
        self.assert_url(
            self._get_request(lib.RequestTarget(hostname)).url,
            hostname,
            self.action,
        )

    def test_url_with_port(self):
        hostname = "host"
        port = 1234
        self.assert_url(
            self._get_request(
                lib.RequestTarget(
                    hostname, dest_list=[Destination(hostname, port)]
                )
            ).url,
            hostname, self.action, port=port,
        )

    def test_url_ipv6(self):
        host_addr = "::1"
        self.assert_url(
            self._get_request(lib.RequestTarget(host_addr)).url,
            "[{0}]".format(host_addr), self.action,
        )

    def test_url_multiaddr(self):
        hosts = ["ring0", "ring1"]
        action = "action"
        request = self._get_request(lib.RequestTarget(
            "label", dest_list=_addr_list_to_dest(hosts)
        ))
        self.assert_url(request.url, hosts[0], action)
        request.next_dest()
        self.assert_url(request.url, hosts[1], action)


class RequestHostTest(TestCase):
    action = "action"

    def _get_request(self, target):
        return lib.Request(target, lib.RequestData(self.action))

    def test_one_host(self):
        hostname = "host"
        request = self._get_request(lib.RequestTarget(hostname))
        self.assertEqual(Destination(hostname, PORT), request.dest)
        self.assertRaises(StopIteration, request.next_dest)

    def test_multiple_hosts(self):
        hosts = ["host1", "host2", "host3"]
        request = self._get_request(lib.RequestTarget(
            "label", dest_list=_addr_list_to_dest(hosts)
        ))
        for hostname in hosts:
            self.assertEqual(Destination(hostname, None), request.dest)
            if hostname == hosts[-1]:
                self.assertRaises(StopIteration, request.next_dest)
            else:
                request.next_dest()


class RequestCookiesTest(TestCase):
    @staticmethod
    def _get_request(token=None):
        return lib.Request(
            lib.RequestTarget("host", token=token), lib.RequestData("action")
        )

    def test_with_token(self):
        token = "token1"
        self.assertEqual({"token": token}, self._get_request(token).cookies)

    def test_without_token(self):
        self.assertEqual({}, self._get_request().cookies)


class ResponseTest(TestCase):
    @staticmethod
    def fixture_handle(info, request, data, debug):
        handle = MockCurl(info)
        handle.request_obj = request
        handle.output_buffer = io.BytesIO()
        handle.output_buffer.write(data.encode("utf-8"))
        handle.debug_buffer = io.BytesIO()
        handle.debug_buffer.write(debug.encode("utf-8"))
        return handle

    def test_connection_successful(self):
        request = lib.Request(
            lib.RequestTarget("host"), lib.RequestData("request")
        )
        output = "output"
        debug = "debug"
        response_code = 200
        handle = self.fixture_handle(
            {pycurl.RESPONSE_CODE: 200}, request, output, debug
        )
        response = lib.Response.connection_successful(handle)
        self.assertEqual(request, response.request)
        self.assertTrue(response.was_connected)
        self.assertIsNone(response.errno)
        self.assertIsNone(response.error_msg)
        self.assertEqual(output, response.data)
        self.assertEqual(debug, response.debug)
        self.assertEqual(response_code, response.response_code)

    def test_connection_failure(self):
        request = lib.Request(
            lib.RequestTarget("host"), lib.RequestData("request")
        )
        output = "output"
        debug = "debug"
        errno = 1
        error_msg = "error"
        handle = self.fixture_handle({}, request, output, debug)
        response = lib.Response.connection_failure(handle, errno, error_msg)
        self.assertEqual(request, response.request)
        self.assertFalse(response.was_connected)
        self.assertEqual(errno, response.errno)
        self.assertEqual(error_msg, response.error_msg)
        self.assertEqual(output, response.data)
        self.assertEqual(debug, response.debug)
        self.assertIsNone(response.response_code)


@mock.patch("pcs.common.node_communicator.pycurl.Curl")
class CreateRequestHandleTest(TestCase):
    # pylint: disable=no-member, protected-access
    _common_opts = {
        pycurl.PROTOCOLS: pycurl.PROTO_HTTPS,
        pycurl.VERBOSE: 1,
        pycurl.SSL_VERIFYHOST: 0,
        pycurl.SSL_VERIFYPEER: 0,
        pycurl.NOSIGNAL: 1,
    }

    def test_all_info(self, mock_curl):
        mock_curl.return_value = MockCurl(
            None, b"output", [
                (pycurl.DEBUG_TEXT, b"debug"),
                (pycurl.DEBUG_DATA_OUT, b"info\n"),
            ]
        )
        request = lib.Request(
            lib.RequestTarget(
                "label",
                token="token_val",
                dest_list=_addr_list_to_dest(["host1", "host2"], port=123),
            ),
            lib.RequestData("action", [("data", "value")])
        )
        cookies = {
            "name1": "val1",
            "name2": "val2",
        }
        handle = lib._create_request_handle(request, cookies, 1)
        expected_opts = {
            pycurl.TIMEOUT: 1,
            pycurl.URL: request.url.encode("utf-8"),
            pycurl.COOKIE: "name1=val1;name2=val2;token=token_val".encode(
                "utf-8"
            ),
            pycurl.COPYPOSTFIELDS: "data=value".encode("utf-8"),
        }
        expected_opts.update(self._common_opts)
        self.assertLessEqual(
            set(expected_opts.items()), set(handle.opts.items())
        )
        self.assertIs(request, handle.request_obj)
        self.assertEqual("", handle.output_buffer.getvalue().decode("utf-8"))
        self.assertEqual("", handle.debug_buffer.getvalue().decode("utf-8"))
        handle.perform()
        self.assertEqual(
            "output", handle.output_buffer.getvalue().decode("utf-8")
        )
        self.assertEqual(
            "* debug\n>> info\n", handle.debug_buffer.getvalue().decode("utf-8")
        )

    def test_basic(self, mock_curl):
        mock_curl.return_value = MockCurl(None)
        request = lib.Request(
            lib.RequestTarget("label"), lib.RequestData("action")
        )
        handle = lib._create_request_handle(request, {}, 10)
        expected_opts = {
            pycurl.TIMEOUT: 10,
            pycurl.URL: request.url.encode("utf-8"),
        }
        expected_opts.update(self._common_opts)
        self.assertLessEqual(
            set(expected_opts.items()), set(handle.opts.items())
        )
        self.assertFalse(pycurl.COOKIE in handle.opts)
        self.assertFalse(pycurl.COPYPOSTFIELDS in handle.opts)
        self.assertIs(request, handle.request_obj)
        self.assertEqual("", handle.output_buffer.getvalue().decode("utf-8"))
        self.assertEqual("", handle.debug_buffer.getvalue().decode("utf-8"))
        handle.perform()
        self.assertEqual("", handle.output_buffer.getvalue().decode("utf-8"))
        self.assertEqual("", handle.debug_buffer.getvalue().decode("utf-8"))


def fixture_request(host_id=1, action="action"):
    return lib.Request(
        lib.RequestTarget("host{0}".format(host_id)), lib.RequestData(action),
    )


class CommunicatorBaseTest(TestCase):
    def setUp(self):
        self.mock_com_log = mock.MagicMock(
            spec_set=lib.CommunicatorLoggerInterface
        )

    def get_communicator(self):
        return lib.Communicator(self.mock_com_log, None, None)

    def get_multiaddress_communicator(self):
        return lib.MultiaddressCommunicator(self.mock_com_log, None, None)


@mock.patch(
    "pcs.common.node_communicator.pycurl.CurlMulti",
    side_effect=lambda: MockCurlMulti([1])
)
@mock.patch("pcs.common.node_communicator._create_request_handle")
class CommunicatorSimpleTest(CommunicatorBaseTest):
    def get_response(self, com, mock_create_handle, handle):
        request = fixture_request(0, "action")
        handle.request_obj = request
        mock_create_handle.return_value = handle
        com.add_requests([request])
        self.assertEqual(0, self.mock_com_log.log_request_start.call_count)
        response_list = list(com.start_loop())
        self.assertEqual(1, len(response_list))
        response = response_list[0]
        self.assertIs(handle, response.handle)
        self.assertIs(request, response.request)
        mock_create_handle.assert_called_once_with(
            request, {}, settings.default_request_timeout
        )
        return response

    def assert_common_checks(self, com, response):
        # pylint: disable=protected-access
        self.assertEqual(response.handle.error is None, response.was_connected)
        self.mock_com_log.log_request_start.assert_called_once_with(
            response.request
        )
        self.mock_com_log.log_response.assert_called_once_with(response)
        self.assertEqual(0, self.mock_com_log.log_retry.call_count)
        self.assertEqual(0, self.mock_com_log.log_no_more_addresses.call_count)
        com._multi_handle.assert_no_handle_left()

    def test_simple(self, mock_create_handle, _):
        com = self.get_communicator()
        response = self.get_response(com, mock_create_handle, MockCurl())
        self.assert_common_checks(com, response)

    def test_failure(self, mock_create_handle, _):
        com = self.get_communicator()
        expected_reason = "expected reason"
        errno = pycurl.E_SEND_ERROR
        response = self.get_response(
            com, mock_create_handle, MockCurl(error=(errno, expected_reason))
        )
        self.assert_common_checks(com, response)
        self.assertEqual(errno, response.errno)
        self.assertEqual(expected_reason, response.error_msg)


class CommunicatorMultiTest(CommunicatorBaseTest):
    @mock.patch("pcs.common.node_communicator._create_request_handle")
    @mock.patch(
        "pcs.common.node_communicator.pycurl.CurlMulti",
        side_effect=lambda: MockCurlMulti([1, 1])
    )
    def test_call_start_loop_multiple_times(self, _, mock_create_handle):
        com = self.get_communicator()
        mock_create_handle.side_effect = lambda request, _, __: MockCurl(
            request=request
        )
        com.add_requests([fixture_request(i) for i in range(2)])
        next(com.start_loop())
        with self.assertRaises(AssertionError):
            next(com.start_loop())

    @mock.patch("pcs.common.node_communicator.pycurl.Curl")
    @mock.patch(
        "pcs.common.node_communicator.pycurl.CurlMulti",
        side_effect=lambda: MockCurlMulti([2, 0, 0, 1, 0, 1, 1])
    )
    def test_multiple(self, _, mock_curl):
        com = self.get_communicator()
        action = "action"
        counter = {"counter": 0}
        def _create_mock_curl():
            counter["counter"] += 1
            return (
                MockCurl()
                if counter["counter"] != 2
                else MockCurl(error=(pycurl.E_SEND_ERROR, "reason"))
            )
        mock_curl.side_effect = _create_mock_curl
        request_list = [fixture_request(i, action) for i in range(3)]
        com.add_requests(request_list)
        self.assertEqual(0, self.mock_com_log.log_request_start.call_count)
        response_list = []
        for response in com.start_loop():
            if not response_list:
                request = fixture_request(3, action)
                request_list.append(request)
                com.add_requests([request])
            elif len(response_list) == 3:
                request = fixture_request(4, action)
                request_list.append(request)
                com.add_requests([request])
            response_list.append(response)
        self.assertEqual(len(request_list), len(response_list))
        self.assertEqual(request_list, [r.request for r in response_list])
        for i in range(len(request_list)):
            self.assertEqual(i != 1, response_list[i].was_connected)
        logger_calls = (
            [mock.call.log_request_start(request_list[i]) for i in range(3)]
            +
            [
                mock.call.log_response(response_list[0]),
                mock.call.log_request_start(request_list[3]),
            ]
            +
            [mock.call.log_response(response_list[i]) for i in range(1, 4)]
            +
            [
                mock.call.log_request_start(request_list[4]),
                mock.call.log_response(response_list[4]),
            ]
        )
        self.assertEqual(logger_calls, self.mock_com_log.mock_calls)
        # pylint: disable=no-member, protected-access
        com._multi_handle.assert_no_handle_left()


def fixture_logger_request_retry_calls(response, hostname):
    return [
        mock.call.log_request_start(response.request),
        mock.call.log_response(response),
        mock.call.log_retry(response, hostname),
    ]


@mock.patch.object(lib.Response, "connection_failure")
@mock.patch.object(lib.Response, "connection_successful")
@mock.patch(
    "pcs.common.node_communicator.pycurl.CurlMulti",
    side_effect=lambda: MockCurlMulti([1, 0, 1, 1, 1])
)
@mock.patch("pcs.common.node_communicator._create_request_handle")
class MultiaddressCommunicatorTest(CommunicatorBaseTest):
    def test_success(
        self, mock_create_handle, _, mock_con_successful, mock_con_failure
    ):
        com = self.get_multiaddress_communicator()
        counter = {"counter": 0}
        expected_response_list = []
        def _con_successful(handle):
            response = lib.Response(handle, True)
            expected_response_list.append(response)
            return response

        def _con_failure(handle, errno, err_msg):
            response = lib.Response(handle, False, errno, err_msg)
            expected_response_list.append(response)
            return response

        def _mock_create_request_handle(request, _, __):
            counter["counter"] += 1
            return(
                MockCurl(request=request)
                if counter["counter"] > 2
                else MockCurl(
                    error=(pycurl.E_SEND_ERROR, "reason"),
                    request=request,
                )
            )
        mock_con_successful.side_effect = _con_successful
        mock_con_failure.side_effect = _con_failure
        mock_create_handle.side_effect = _mock_create_request_handle
        request = lib.Request(
            lib.RequestTarget(
                "label",
                dest_list=_addr_list_to_dest(
                    ["host{0}".format(i) for i in range(4)]
                )
            ),
            lib.RequestData("action")
        )
        com.add_requests([request])
        response_list = list(com.start_loop())
        self.assertEqual(1, len(response_list))
        response = response_list[0]
        self.assertIs(response, expected_response_list[-1])
        self.assertTrue(response.was_connected)
        self.assertIs(request, response.request)
        self.assertEqual(Destination("host2", None), request.dest)
        self.assertEqual(3, mock_create_handle.call_count)
        self.assertEqual(3, len(expected_response_list))
        mock_create_handle.assert_has_calls([
            mock.call(request, {}, settings.default_request_timeout)
            for _ in range(3)
        ])
        logger_calls = (
            fixture_logger_request_retry_calls(
                expected_response_list[0], Destination("host0", None)
            )
            +
            fixture_logger_request_retry_calls(
                expected_response_list[1], Destination("host1", None)
            )
            +
            [
                mock.call.log_request_start(request),
                mock.call.log_response(response),
            ]
        )
        self.assertEqual(logger_calls, self.mock_com_log.mock_calls)
        # pylint: disable=no-member, protected-access
        com._multi_handle.assert_no_handle_left()

    def test_failure(
        self, mock_create_handle, _, mock_con_successful, mock_con_failure
    ):
        expected_response_list = []
        def _con_failure(handle, errno, err_msg):
            response = lib.Response(handle, False, errno, err_msg)
            expected_response_list.append(response)
            return response

        mock_con_failure.side_effect = _con_failure
        com = self.get_multiaddress_communicator()
        mock_create_handle.side_effect = lambda request, _, __: MockCurl(
            error=(pycurl.E_SEND_ERROR, "reason"), request=request,
        )
        request = lib.Request(
            lib.RequestTarget(
                "label",
                dest_list=_addr_list_to_dest(
                    ["host{0}".format(i) for i in range(4)]
                ),
            ),
            lib.RequestData("action")
        )
        com.add_requests([request])
        response_list = list(com.start_loop())
        self.assertEqual(1, len(response_list))
        response = response_list[0]
        self.assertFalse(response.was_connected)
        self.assertIs(request, response.request)
        self.assertEqual(Destination("host3", None), request.dest)
        self.assertEqual(4, mock_create_handle.call_count)
        mock_con_successful.assert_not_called()
        self.assertEqual(4, len(expected_response_list))
        mock_create_handle.assert_has_calls([
            mock.call(request, {}, settings.default_request_timeout)
            for _ in range(3)
        ])
        logger_calls = (
            fixture_logger_request_retry_calls(
                expected_response_list[0], Destination("host0", None)
            )
            +
            fixture_logger_request_retry_calls(
                expected_response_list[1], Destination("host1", None)
            )
            +
            fixture_logger_request_retry_calls(
                expected_response_list[2], Destination("host2", None)
            )
            +
            [
                mock.call.log_request_start(request),
                mock.call.log_response(response),
                mock.call.log_no_more_addresses(response)
            ]
        )
        self.assertEqual(logger_calls, self.mock_com_log.mock_calls)
        # pylint: disable=no-member, protected-access
        com._multi_handle.assert_no_handle_left()



from unittest import TestCase
from lxml import etree

from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.xml import etree_to_str

from pcs.lib.cib import nvpair
from pcs.lib.cib.tools import IdProvider

# pylint: disable=no-self-use, protected-access

class AppendNewNvpair(TestCase):
    def test_append_new_nvpair_to_given_element(self):
        nvset_element = etree.fromstring('<nvset id="a"/>')
        id_provider = IdProvider(nvset_element)
        nvpair._append_new_nvpair(nvset_element, "b", "c", id_provider)
        assert_xml_equal(
            etree_to_str(nvset_element),
            """
            <nvset id="a">
                <nvpair id="a-b" name="b" value="c"></nvpair>
            </nvset>
            """
        )

    def test_with_id_provider(self):
        nvset_element = etree.fromstring('<nvset id="a"/>')
        provider = IdProvider(nvset_element)
        provider.book_ids("a-b")
        nvpair._append_new_nvpair(nvset_element, "b", "c", provider)
        assert_xml_equal(
            etree_to_str(nvset_element),
            """
            <nvset id="a">
                <nvpair id="a-b-1" name="b" value="c"></nvpair>
            </nvset>
            """
        )


class UpdateNvsetTest(TestCase):
    def test_updates_nvset(self):
        nvset_element = etree.fromstring("""
            <instance_attributes id="iattrs">
                <nvpair id="iattrs-a" name="a" value="b"/>
                <nvpair id="iattrs-c" name="c" value="d"/>
                <nvpair id="iattrs-e" name="e" value="f"/>
            </instance_attributes>
        """)
        id_provider = IdProvider(nvset_element)
        nvpair.update_nvset(
            nvset_element,
            {
                "a": "B",
                "c": "",
                "g": "h",
            },
            id_provider
        )
        assert_xml_equal(
            """
            <instance_attributes id="iattrs">
                <nvpair id="iattrs-a" name="a" value="B"/>
                <nvpair id="iattrs-e" name="e" value="f"/>
                <nvpair id="iattrs-g" name="g" value="h"/>
            </instance_attributes>
            """,
            etree_to_str(nvset_element)
        )

    def test_empty_value_has_no_effect(self):
        xml = """
            <instance_attributes id="iattrs">
                <nvpair id="iattrs-b" name="a" value="b"/>
                <nvpair id="iattrs-d" name="c" value="d"/>
                <nvpair id="iattrs-f" name="e" value="f"/>
            </instance_attributes>
        """
        nvset_element = etree.fromstring(xml)
        id_provider = IdProvider(nvset_element)
        nvpair.update_nvset(nvset_element, {}, id_provider)
        assert_xml_equal(xml, etree_to_str(nvset_element))

    def test_keep_empty_nvset(self):
        xml_pre = """
            <resource>
                <instance_attributes id="iattrs">
                    <nvpair id="iattrs-a" name="a" value="b"/>
                </instance_attributes>
            </resource>
        """
        xml_post = """
            <resource>
                <instance_attributes id="iattrs" />
            </resource>
        """
        xml = etree.fromstring(xml_pre)
        nvset_element = xml.find("instance_attributes")
        id_provider = IdProvider(nvset_element)
        nvpair.update_nvset(nvset_element, {"a": ""}, id_provider)
        assert_xml_equal(xml_post, etree_to_str(xml))


class SetNvpairInNvsetTest(TestCase):
    def setUp(self):
        self.nvset = etree.Element("nvset", id="nvset")
        etree.SubElement(
            self.nvset, "nvpair", id="nvset-attr", name="attr", value="1"
        )
        etree.SubElement(
            self.nvset, "nvpair", id="nvset-attr2", name="attr2", value="2"
        )
        etree.SubElement(
            self.nvset, "notnvpair", id="nvset-test", name="test", value="0"
        )
        self.id_provider = IdProvider(self.nvset)

    def test_update(self):
        nvpair.set_nvpair_in_nvset(self.nvset, "attr", "10", self.id_provider)
        assert_xml_equal(
            """
            <nvset id="nvset">
                <nvpair id="nvset-attr" name="attr" value="10"/>
                <nvpair id="nvset-attr2" name="attr2" value="2"/>
                <notnvpair id="nvset-test" name="test" value="0"/>
            </nvset>
            """,
            etree_to_str(self.nvset)
        )

    def test_add(self):
        nvpair.set_nvpair_in_nvset(self.nvset, "test", "0", self.id_provider)
        assert_xml_equal(
            """
            <nvset id="nvset">
                <nvpair id="nvset-attr" name="attr" value="1"/>
                <nvpair id="nvset-attr2" name="attr2" value="2"/>
                <notnvpair id="nvset-test" name="test" value="0"/>
                <nvpair id="nvset-test-1" name="test" value="0"/>
            </nvset>
            """,
            etree_to_str(self.nvset)
        )

    def test_remove(self):
        nvpair.set_nvpair_in_nvset(self.nvset, "attr2", "", self.id_provider)
        assert_xml_equal(
            """
            <nvset id="nvset">
                <nvpair id="nvset-attr" name="attr" value="1"/>
                <notnvpair id="nvset-test" name="test" value="0"/>
            </nvset>
            """,
            etree_to_str(self.nvset)
        )

    def test_remove_not_existing(self):
        nvpair.set_nvpair_in_nvset(self.nvset, "attr3", "", self.id_provider)
        assert_xml_equal(
            """
            <nvset id="nvset">
                <nvpair id="nvset-attr" name="attr" value="1"/>
                <nvpair id="nvset-attr2" name="attr2" value="2"/>
                <notnvpair id="nvset-test" name="test" value="0"/>
            </nvset>
            """,
            etree_to_str(self.nvset)
        )

class AppendNewNvsetTest(TestCase):
    def test_append_new_nvset_to_given_element(self):
        context_element = etree.fromstring('<context id="a"/>')
        id_provider = IdProvider(context_element)
        nvpair.append_new_nvset(
            "instance_attributes",
            context_element,
            {
                "a": "b",
                "c": "d",
            },
            id_provider
        )
        assert_xml_equal(
            """
                <context id="a">
                    <instance_attributes id="a-instance_attributes">
                        <nvpair
                            id="a-instance_attributes-a" name="a" value="b"
                        />
                        <nvpair
                            id="a-instance_attributes-c" name="c" value="d"
                        />
                    </instance_attributes>
                </context>
            """,
            etree_to_str(context_element)
        )

    def test_with_id_provider_booked_ids(self):
        context_element = etree.fromstring('<context id="a"/>')
        provider = IdProvider(context_element)
        provider.book_ids("a-instance_attributes", "a-instance_attributes-1-a")
        nvpair.append_new_nvset(
            "instance_attributes",
            context_element,
            {
                "a": "b",
                "c": "d",
            },
            provider
        )
        assert_xml_equal(
            """
                <context id="a">
                    <instance_attributes id="a-instance_attributes-1">
                        <nvpair
                            id="a-instance_attributes-1-a-1" name="a" value="b"
                        />
                        <nvpair
                            id="a-instance_attributes-1-c" name="c" value="d"
                        />
                    </instance_attributes>
                </context>
            """,
            etree_to_str(context_element)
        )


class ArrangeFirstNvsetTest(TestCase):
    def setUp(self):
        self.root = etree.Element("root", id="root")
        self.nvset = etree.SubElement(self.root, "nvset", id="nvset")
        etree.SubElement(
            self.nvset, "nvpair", id="nvset-attr", name="attr", value="1"
        )
        etree.SubElement(
            self.nvset, "nvpair", id="nvset-attr2", name="attr2", value="2"
        )
        etree.SubElement(
            self.nvset, "notnvpair", id="nvset-test", name="test", value="0"
        )
        self.id_provider = IdProvider(self.nvset)

    def test_empty_value_has_no_effect(self):
        nvpair.arrange_first_nvset("nvset", self.root, {}, self.id_provider)
        assert_xml_equal(
            """
                <nvset id="nvset">
                    <nvpair id="nvset-attr" name="attr" value="1"/>
                    <nvpair id="nvset-attr2" name="attr2" value="2"/>
                    <notnvpair id="nvset-test" name="test" value="0"/>
                </nvset>
            """,
            etree_to_str(self.nvset)
        )

    def test_update_existing_nvset(self):
        nvpair.arrange_first_nvset(
            "nvset",
            self.root,
            {
                "attr": "10",
                "new_one": "20",
                "test": "0",
                "attr2": ""
            },
            self.id_provider
        )
        assert_xml_equal(
            """
                <nvset id="nvset">
                    <nvpair id="nvset-attr" name="attr" value="10"/>
                    <notnvpair id="nvset-test" name="test" value="0"/>
                    <nvpair id="nvset-new_one" name="new_one" value="20"/>
                    <nvpair id="nvset-test-1" name="test" value="0"/>
                </nvset>
            """,
            etree_to_str(self.nvset)
        )

    def test_create_new_nvset_if_does_not_exist(self):
        root = etree.Element("root", id="root")
        nvpair.arrange_first_nvset(
            "nvset",
            root,
            {
                "attr": "10",
                "new_one": "20",
                "test": "0",
                "attr2": ""
            },
            self.id_provider
        )

        assert_xml_equal(
            """
            <root id="root">
                <nvset id="root-nvset">
                    <nvpair id="root-nvset-attr" name="attr" value="10"/>
                    <nvpair id="root-nvset-new_one" name="new_one" value="20"/>
                    <nvpair id="root-nvset-test" name="test" value="0"/>
                </nvset>
            </root>
            """,
            etree_to_str(root)
        )


class GetNvsetTest(TestCase):
    def test_success(self):
        nvset = etree.XML(
            """
            <nvset>
                <nvpair id="nvset-name1" name="name1" value="value1"/>
                <nvpair id="nvset-name2" name="name2" value="value2"/>
                <nvpair id="nvset-name3" name="name3"/>
            </nvset>
            """
        )
        self.assertEqual(
            [
                {
                    "id": "nvset-name1",
                    "name": "name1",
                    "value": "value1"
                },
                {
                    "id": "nvset-name2",
                    "name": "name2",
                    "value": "value2"
                },
                {
                    "id": "nvset-name3",
                    "name": "name3",
                    "value": ""
                }
            ],
            nvpair.get_nvset(nvset)
        )

class GetValue(TestCase):
    def assert_find_value(self, tag_name, name, value, xml, default=None):
        self.assertEqual(
            value,
            nvpair.get_value(tag_name, etree.fromstring(xml), name, default)
        )

    def test_return_value_when_name_exists(self):
        self.assert_find_value(
            "meta_attributes",
            "SOME-NAME",
            "some-value",
            """
                <context>
                    <meta_attributes>
                        <nvpair name="SOME-NAME" value="some-value" />
                        <nvpair name="OTHER-NAME" value="other-value" />
                    </meta_attributes>
                </context>
            """,
        )

    def test_return_none_when_name_not_exists(self):
        self.assert_find_value(
            "instance_attributes",
            "SOME-NAME",
            value=None,
            xml="""
                <context>
                    <instance_attributes>
                        <nvpair name="another-name" value="some-value" />
                    </instance_attributes>
                </context>
            """,
        )

    def test_return_default_when_name_not_exists(self):
        self.assert_find_value(
            "instance_attributes",
            "SOME-NAME",
            value="DEFAULT",
            xml="""
                <context>
                    <instance_attributes>
                        <nvpair name="another-name" value="some-value" />
                    </instance_attributes>
                </context>
            """,
            default="DEFAULT",
        )

    def test_return_none_when_no_nvpair(self):
        self.assert_find_value(
            "instance_attributes",
            "SOME-NAME",
            value=None,
            xml="""
                <context>
                    <instance_attributes />
                </context>
            """,
        )

    def test_return_none_when_no_nvset(self):
        self.assert_find_value(
            "instance_attributes",
            "SOME-NAME",
            value=None,
            xml="""
                <context>
                </context>
            """,
        )

class GetNvsetAsDictTest(TestCase):
    def test_no_element(self):
        resource_element = etree.fromstring("<primitive/>")
        self.assertEqual(
            dict(),
            nvpair.get_nvset_as_dict("meta_attributes", resource_element),
        )

    def test_empty(self):
        resource_element = etree.fromstring("""
            <primitive>
                <meta_attributes/>
            </primitive>
        """)
        self.assertEqual(
            dict(),
            nvpair.get_nvset_as_dict("meta_attributes", resource_element),
        )

    def test_non_empty(self):
        resource_element = etree.fromstring("""
            <primitive>
                <meta_attributes>
                    <nvpair id="a" name="attr_name" value="value"/>
                    <nvpair id="b" name="other_name" value="other-value"/>
                </meta_attributes>
            </primitive>
        """)
        self.assertEqual(
            dict(
                attr_name="value",
                other_name="other-value",
            ),
            nvpair.get_nvset_as_dict("meta_attributes", resource_element),
        )

    def test_multiple_nvsets(self):
        resource_element = etree.fromstring("""
            <primitive>
                <meta_attributes>
                    <nvpair id="a" name="attr_name" value="value"/>
                    <nvpair id="b" name="other_name" value="other-value"/>
                </meta_attributes>
                <meta_attributes>
                    <nvpair id="a" name="attr_name2" value="value2"/>
                    <nvpair id="b" name="other_name2" value="other-value2"/>
                </meta_attributes>
            </primitive>
        """)
        self.assertEqual(
            dict(
                attr_name="value",
                other_name="other-value",
            ),
            nvpair.get_nvset_as_dict("meta_attributes", resource_element),
        )

class HasMetaAttribute(TestCase):
    def test_return_false_if_does_not_have_such_attribute(self):
        resource_element = etree.fromstring("""<primitive/>""")
        self.assertFalse(
            nvpair.has_meta_attribute(resource_element, "attr_name")
        )

    def test_return_true_if_such_meta_attribute_exists(self):
        resource_element = etree.fromstring("""
            <primitive>
                <meta_attributes>
                    <nvpair id="a" name="attr_name" value="value"/>
                    <nvpair id="b" name="other_name" value="other-value"/>
                </meta_attributes>
            </primitive>
        """)
        self.assertTrue(
            nvpair.has_meta_attribute(resource_element, "attr_name")
        )

    def test_return_false_if_meta_attribute_exists_but_in_nested_element(self):
        resource_element = etree.fromstring("""
            <group>
                <primitive>
                    <meta_attributes>
                        <nvpair id="a" name="attr_name" value="value"/>
                    </meta_attributes>
                </primitive>
            </group>
        """)
        self.assertFalse(
            nvpair.has_meta_attribute(resource_element, "attr_name")
        )



import shutil
from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource as rc,
    is_minimum_pacemaker_version,
)
from pcs_test.tools.pcs_runner import PcsRunner

PCMK_2_0_3_PLUS = is_minimum_pacemaker_version(2, 0, 3)

class StonithWarningTest(TestCase, AssertPcsMixin):
    empty_cib = rc("cib-empty.xml")
    temp_cib = rc("temp-cib.xml")
    corosync_conf = rc("corosync.conf")

    def setUp(self):
        shutil.copy(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib)

    def fixture_stonith_action(self):
        self.assert_pcs_success(
            "stonith create Sa fence_apc ip=i username=u action=reboot --force",
            "Warning: stonith option 'action' is deprecated and should not be"
                " used, use pcmk_off_action, pcmk_reboot_action instead\n"
        )

    def fixture_stonith_cycle(self):
        self.assert_pcs_success(
            "stonith create Sc fence_ipmilan method=cycle"
        )

    def fixture_resource(self):
        self.assert_pcs_success(
            "resource create dummy ocf:pacemaker:Dummy action=reboot "
                "method=cycle --force"
            ,
            "Warning: invalid resource options: 'action', 'method', allowed "
                "options are: 'envfile', 'fail_start_on', 'fake', 'op_sleep', "
                "'passwd', 'state', 'trace_file', 'trace_ra'\n"
        )

    def test_warning_stonith_action(self):
        self.fixture_stonith_action()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'

                    Cluster Summary:
                """)
            )
        else:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'

                    Stack: unknown
                    Current DC: NONE
                """)
            )

    def test_warning_stonith_method_cycle(self):
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Cluster Summary:
                """)
            )
        else:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Stack: unknown
                    Current DC: NONE
                """)
            )

    def test_stonith_warnings(self):
        self.fixture_stonith_action()
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Cluster Summary:
                """)
            )
        else:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Stack: unknown
                    Current DC: NONE
                """)
            )

    def test_warn_when_no_stonith(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    No stonith devices and stonith-enabled is not false

                    Cluster Summary:
                """)
            )
        else:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99

                    WARNINGS:
                    No stonith devices and stonith-enabled is not false

                    Stack: unknown
                    Current DC: NONE
                """)
            )

    def test_disabled_stonith_does_not_care_about_missing_devices(self):
        self.assert_pcs_success("property set stonith-enabled=false")
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99
                    Cluster Summary:
                """)
            )
        else:
            self.assert_pcs_success(
                "status",
                stdout_start=dedent("""\
                    Cluster name: test99
                    Stack: unknown
                    Current DC: NONE
                """)
            )



from pcs import usage
from pcs.cli.booth import command
from pcs.cli.common.routing import create_router
from pcs.resource import resource_remove, resource_restart


booth_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.booth(argv),
        "config": command.config_show,
        "setup": command.config_setup,
        "destroy": command.config_destroy,
        "ticket": create_router(
            {
                "help": lambda lib, argv, modifiers: usage.booth(["ticket"]),
                "add": command.config_ticket_add,
                "delete": command.config_ticket_remove,
                "remove": command.config_ticket_remove,
                "grant": command.ticket_grant,
                "revoke": command.ticket_revoke,
            },
            ["booth", "ticket"]
        ),
        "create": command.create_in_cluster,
        "delete": command.get_remove_from_cluster(resource_remove),
        "remove": command.get_remove_from_cluster(resource_remove),
        "restart": command.get_restart(resource_restart),
        "sync": command.sync,
        "pull": command.pull,
        "enable": command.enable,
        "disable": command.disable,
        "start": command.start,
        "stop": command.stop,
        "status": command.status,
    },
    ["booth"]
)



import os
from unittest import mock, TestCase

from OpenSSL import SSL

from pcs_test.tools.misc import get_test_resource as rc

from pcs.daemon.ssl import PcsdSSL, CertKeyPair, SSLCertKeyException

SERVER_NAME = "pcsd-daemon"
SSL_OPTIONS = 0
SSL_CIPHERS = "DEFAULT:!RC4"
CERT = rc("daemon.cert")
KEY = rc("daemon.key")

def remove_ssl_files():
    if os.path.exists(CERT):
        os.remove(CERT)
    if os.path.exists(KEY):
        os.remove(KEY)

def damage_ssl_files():
    with open(CERT, "w") as cert:
        cert.write("bad content")
    with open(KEY, "w") as key:
        key.write("bad content")

# various versions of OpenSSL / PyOpenSSL emit different messages
DAMAGED_SSL_FILES_ERRORS_1 = (
    f"Invalid SSL certificate '{CERT}':"
        " 'PEM routines:PEM_read_bio:no start line'"
    ,
    f"Invalid SSL key '{KEY}': 'PEM routines:PEM_read_bio:no start line'"
)
DAMAGED_SSL_FILES_ERRORS_2 = (
    f"Invalid SSL certificate '{CERT}':"
        " 'PEM routines:get_name:no start line'"
    ,
    f"Invalid SSL key '{KEY}': 'PEM routines:get_name:no start line'"
)

class Pair(TestCase):
    def setUp(self):
        remove_ssl_files()
        self.pair = CertKeyPair(CERT, KEY)

    def tearDown(self):
        remove_ssl_files()

    def test_generated_key_check_success(self):
        self.pair.regenerate(SERVER_NAME)
        self.assertEqual(self.pair.check(), [])

    def test_error_if_files_with_bad_content(self):
        damage_ssl_files()
        self.assertTrue(
            self.pair.check()
            in
            [
                list(DAMAGED_SSL_FILES_ERRORS_1),
                list(DAMAGED_SSL_FILES_ERRORS_2),
            ]
        )

    @mock.patch("pcs.daemon.ssl.SSL.Context.use_privatekey")
    def test_error_if_short_key(self, mock_use_key):
        mock_use_key.side_effect = SSL.Error("reason")
        self.pair.regenerate(SERVER_NAME, 1024)
        errors = self.pair.check()
        self.assertEqual(
            errors,
            ["Unable to load SSL certificate and/or key: reason"]
        )

    def test_error_if_cert_does_not_match_key(self):
        self.pair.regenerate(SERVER_NAME)
        with open(KEY, "rb") as key_file:
            key_content = key_file.read()
        self.pair.regenerate(SERVER_NAME)
        with open(KEY, "wb") as key_file:
            key_file.write(key_content)

        errors = self.pair.check()
        self.assertEqual(len(errors), 1)
        self.assertTrue(
            errors[0].startswith("SSL certificate does not match the key:")
        )

class PcsdSSLTest(TestCase):
    def setUp(self):
        remove_ssl_files()
        self.pcsd_ssl = PcsdSSL(
            SERVER_NAME,
            CERT,
            KEY,
            SSL_OPTIONS,
            SSL_CIPHERS
        )

    def tearDown(self):
        remove_ssl_files()

    def test_is_ok_when_ssl_files_does_not_exist(self):
        self.pcsd_ssl.guarantee_valid_certs()

    def test_raises_when_ssl_files_are_damaged(self):
        damage_ssl_files()
        with self.assertRaises(SSLCertKeyException) as ctx_manager:
            self.pcsd_ssl.guarantee_valid_certs()
        self.assertTrue(
            ctx_manager.exception.args
            in
            [DAMAGED_SSL_FILES_ERRORS_1, DAMAGED_SSL_FILES_ERRORS_2]
        )

    def test_context_uses_given_options(self):
        self.pcsd_ssl.guarantee_valid_certs()
        ssl_context = self.pcsd_ssl.create_context()
        self.assertEqual(ssl_context.options, SSL_OPTIONS)



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs.common import report_codes
from pcs.lib.commands import cib_options

FIXTURE_INITIAL_DEFAULTS = """
    <op_defaults>
        <meta_attributes id="op_defaults-options">
            <nvpair id="op_defaults-options-a" name="a" value="b"/>
            <nvpair id="op_defaults-options-b" name="b" value="c"/>
        </meta_attributes>
    </op_defaults>
"""

class SetOperationsDefaults(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def tearDown(self):
        self.env_assist.assert_reports([
            fixture.warn(report_codes.DEFAULTS_CAN_BE_OVERRIDEN)
        ])

    def test_change(self):
        self.config.runner.cib.load(optional_in_conf=FIXTURE_INITIAL_DEFAULTS)
        self.config.env.push_cib(optional_in_conf="""
            <op_defaults>
                <meta_attributes id="op_defaults-options">
                    <nvpair id="op_defaults-options-a" name="a" value="B"/>
                    <nvpair id="op_defaults-options-b" name="b" value="C"/>
                </meta_attributes>
            </op_defaults>
        """)
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {
                "a": "B",
                "b": "C",
            }
        )

    def test_add(self):
        self.config.runner.cib.load(optional_in_conf=FIXTURE_INITIAL_DEFAULTS)
        self.config.env.push_cib(optional_in_conf="""
            <op_defaults>
                <meta_attributes id="op_defaults-options">
                    <nvpair id="op_defaults-options-a" name="a" value="b"/>
                    <nvpair id="op_defaults-options-b" name="b" value="c"/>
                    <nvpair id="op_defaults-options-c" name="c" value="d"/>
                </meta_attributes>
            </op_defaults>
        """)
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {"c": "d"},
        )

    def test_remove(self):
        self.config.runner.cib.load(optional_in_conf=FIXTURE_INITIAL_DEFAULTS)
        self.config.env.push_cib(
            remove=
                "./configuration/op_defaults/meta_attributes/nvpair[@name='a']"
        )
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {"a": ""},
        )

    def test_add_section_if_missing(self):
        self.config.runner.cib.load()
        self.config.env.push_cib(optional_in_conf="""
            <op_defaults>
                <meta_attributes id="op_defaults-options">
                    <nvpair id="op_defaults-options-a" name="a" value="A"/>
                </meta_attributes>
            </op_defaults>
        """)
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {"a": "A",}
        )

    def test_add_meta_if_missing(self):
        self.config.runner.cib.load(optional_in_conf="<op_defaults />")
        self.config.env.push_cib(optional_in_conf="""
            <op_defaults>
                <meta_attributes id="op_defaults-options">
                    <nvpair id="op_defaults-options-a" name="a" value="A"/>
                </meta_attributes>
            </op_defaults>
        """)
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {"a": "A",}
        )

    def test_dont_add_section_if_only_removing(self):
        self.config.runner.cib.load()
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {
                "a": "",
                "b": "",
            }
        )

    def test_dont_add_meta_if_only_removing(self):
        self.config.runner.cib.load(optional_in_conf="<op_defaults />")
        self.config.env.push_cib(optional_in_conf="<op_defaults />")
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {
                "a": "",
                "b": "",
            }
        )

    def test_keep_section_when_empty(self):
        self.config.runner.cib.load(optional_in_conf=FIXTURE_INITIAL_DEFAULTS)
        self.config.env.push_cib(remove="./configuration/op_defaults//nvpair")
        cib_options.set_operations_defaults(
            self.env_assist.get_env(),
            {
                "a": "",
                "b": "",
            }
        )



from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call
)

class SbdShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def enable_sbd(
        self, node_labels=None, communication_list=None,
        name="http.sbd.enable_sbd"
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/sbd_enable"
        )

    def disable_sbd(
        self, node_labels=None, communication_list=None,
        name="http.sbd.disable_sbd"
    ):
        """
        Create a call for disabling sbd on nodes

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/sbd_disable"
        )

    def check_sbd(
        self, communication_list=None, name="http.sbd.check_sbd"
    ):
        place_multinode_call(
            self.__calls,
            name,
            communication_list=communication_list,
            action="remote/check_sbd",
        )

    def set_sbd_config(
        self, config_generator=None, node_labels=None, communication_list=None,
        name="http.sbd.set_sbd_config"
    ):
        if bool(config_generator) == bool(communication_list):
            raise AssertionError(
                "Exactly one of 'config_generator', 'communication_list' "
                "must be specified"
            )
        if config_generator and not node_labels:
            raise AssertionError(
                "'node_labels' has to be defined if 'config_generator' is used"
            )
        if communication_list is None:
            communication_list = [
                dict(
                    param_list=[("config", config_generator(node))],
                    label=node,
                ) for node in node_labels
            ]
        place_multinode_call(
            self.__calls,
            name,
            None,
            communication_list,
            action="remote/set_sbd_config",
        )



from pcs import (
    stonith,
    utils,
)
from pcs.cli.common import parse_args
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.reports import process_library_reports
from pcs.common.tools import indent
from pcs.lib.node import get_existing_nodes_names


def quorum_config_cmd(lib, argv, modifiers):
    """
    Options:
      * --corosync_conf - mocked corosync configuration file
    """
    modifiers.ensure_only_supported("--corosync_conf")
    if argv:
        raise CmdLineInputError()
    config = lib.quorum.get_config()
    print("\n".join(quorum_config_to_str(config)))

def quorum_config_to_str(config):
    """
    Commandline options: no options
    """
    lines = []

    lines.append("Options:")
    if "options" in config and config["options"]:
        lines.extend(indent([
            "{n}: {v}".format(n=name, v=value)
            for name, value in sorted(config["options"].items())
        ]))

    if "device" in config and config["device"]:
        lines.append("Device:")
        lines.extend(indent([
            "{n}: {v}".format(n=name, v=value)
            for name, value in sorted(
                config["device"].get("generic_options", {}).items()
            )
        ]))

        model_settings = [
            "Model: {m}".format(m=config["device"].get("model", ""))
        ]
        model_settings.extend(indent([
            "{n}: {v}".format(n=name, v=value)
            for name, value in sorted(
                config["device"].get("model_options", {}).items()
            )
        ]))
        lines.extend(indent(model_settings))

        heuristics_options = config["device"].get("heuristics_options", {})
        if heuristics_options:
            heuristics_settings = ["Heuristics:"]
            heuristics_settings.extend(indent([
                "{n}: {v}".format(n=name, v=value)
                for name, value in sorted(heuristics_options.items())
            ]))
            lines.extend(indent(heuristics_settings))

    return lines

def quorum_expected_votes_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    lib.quorum.set_expected_votes_live(argv[0])

def quorum_status_cmd(lib, argv, modifiers):
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    print(lib.quorum.status())

def quorum_update_cmd(lib, argv, modifiers):
    """
    Options:
      * --skip-offline - skip offline nodes
      * --force - force changes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--force", "--corosync_conf", "--request-timeout"
    )
    options = parse_args.prepare_options(argv)
    if not options:
        raise CmdLineInputError()

    lib.quorum.set_options(
        options,
        skip_offline_nodes=modifiers.get("--skip-offline"),
        force=modifiers.get("--force"),
    )

def _parse_quorum_device_groups(arg_list):
    keyword_list = ["model", "heuristics"]
    groups = parse_args.group_by_keywords(
        arg_list,
        set(keyword_list),
        implicit_first_group_key="generic",
        keyword_repeat_allowed=False,
        only_found_keywords=True
    )
    for keyword in keyword_list:
        if keyword not in groups:
            continue
        if not groups[keyword]:
            raise CmdLineInputError(
                "No {0} options specified".format(keyword)
            )
    return groups

def quorum_device_add_cmd(lib, argv, modifiers):
    """
    Options:
      * --force - allow unknown model or options
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
      * --corosync_conf - mocked corosync configuration file
    """
    modifiers.ensure_only_supported(
        "--force", "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    groups = _parse_quorum_device_groups(argv)
    model_and_model_options = groups.get("model", [])
    # we expect "model" keyword once, followed by the actual model value
    if not model_and_model_options or "=" in model_and_model_options[0]:
        raise CmdLineInputError()

    generic_options = parse_args.prepare_options(groups.get("generic", []))
    model = model_and_model_options[0]
    model_options = parse_args.prepare_options(model_and_model_options[1:])
    heuristics_options = parse_args.prepare_options(
        groups.get("heuristics", [])
    )

    if "model" in generic_options:
        raise CmdLineInputError("Model cannot be specified in generic options")

    lib.quorum.add_device(
        model,
        model_options,
        generic_options,
        heuristics_options,
        force_model=modifiers.get("--force"),
        force_options=modifiers.get("--force"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )

def quorum_device_remove_cmd(lib, argv, modifiers):
    """
    Options:
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    if argv:
        raise CmdLineInputError()

    lib.quorum.remove_device(
        skip_offline_nodes=modifiers.get("--skip-offline")
    )

def quorum_device_status_cmd(lib, argv, modifiers):
    """
    Options:
      * --full - more detailed output
    """
    modifiers.ensure_only_supported("--full")
    if argv:
        raise CmdLineInputError()
    print(lib.quorum.status_device(modifiers.get("--full")))

def quorum_device_update_cmd(lib, argv, modifiers):
    """
    Options:
      * --force - allow unknown options
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--force", "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    groups = _parse_quorum_device_groups(argv)
    if not groups:
        raise CmdLineInputError()
    generic_options = parse_args.prepare_options(groups.get("generic", []))
    model_options = parse_args.prepare_options(groups.get("model", []))
    heuristics_options = parse_args.prepare_options(
        groups.get("heuristics", [])
    )

    if "model" in generic_options:
        raise CmdLineInputError("Model cannot be specified in generic options")

    lib.quorum.update_device(
        model_options,
        generic_options,
        heuristics_options,
        force_options=modifiers.get("--force"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )

def quorum_device_heuristics_remove_cmd(lib, argv, modifiers):
    """
    Options:
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--corosync_conf", "--request-timeout"
    )
    if argv:
        raise CmdLineInputError()
    lib.quorum.remove_device_heuristics(
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )

# TODO switch to new architecture, move to lib
def quorum_unblock_cmd(lib, argv, modifiers):
    """
    Options:
      * --force - no error when removing non existing property and no warning
        about this action
    """
    modifiers.ensure_only_supported("--force")
    if argv:
        raise CmdLineInputError()

    output, retval = utils.run(
        ["corosync-cmapctl", "-g", "runtime.votequorum.wait_for_all_status"]
    )
    if retval != 0:
        utils.err("unable to check quorum status")
    if output.split("=")[-1].strip() != "1":
        utils.err("cluster is not waiting for nodes to establish quorum")

    all_nodes, report_list = get_existing_nodes_names(
        utils.get_corosync_conf_facade()
    )
    if report_list:
        process_library_reports(report_list)

    unjoined_nodes = set(all_nodes) - set(utils.getCorosyncActiveNodes())
    if not unjoined_nodes:
        utils.err("no unjoined nodes found")
    if not modifiers.get("--force"):
        answer = utils.get_terminal_input(
            (
                "WARNING: If node(s) {nodes} are not powered off or they do"
                + " have access to shared resources, data corruption and/or"
                + " cluster failure may occur. Are you sure you want to"
                + " continue? [y/N] "
            ).format(nodes=", ".join(unjoined_nodes))
        )
        if answer.lower() not in ["y", "yes"]:
            print("Canceled")
            return
    for node in unjoined_nodes:
        # pass --force so no warning will be displayed
        stonith.stonith_confirm(
            lib, [node], parse_args.InputModifiers({"--force": ""})
        )

    output, retval = utils.run(
        ["corosync-cmapctl", "-s", "quorum.cancel_wait_for_all", "u8", "1"]
    )
    if retval != 0:
        utils.err("unable to cancel waiting for nodes")
    print("Quorum unblocked")

    startup_fencing = utils.get_set_properties().get("startup-fencing", "")
    utils.set_cib_property(
        "startup-fencing",
        "false" if startup_fencing.lower() != "false" else "true"
    )
    utils.set_cib_property("startup-fencing", startup_fencing)
    print("Waiting for nodes canceled")



from unittest import TestCase

from pcs.common.tools import xml_fromstring

class XmlFromstring(TestCase):
    def test_large_xml(self):
        # pylint: disable=no-self-use
        #it raises on a huge xml without the flag huge_tree=True
        #see https://bugzilla.redhat.com/show_bug.cgi?id=1506864
        xml_fromstring(large_xml)

large_xml = """
<cib admin_epoch="1" epoch="1305" num_updates="0" validate-with="pacemaker-2.8">
      <configuration>
        <crm_config/>
        <nodes/>
        <resources>{0}</resources>
        <constraints/>
      </configuration>
  <status>{1}</status>
</cib>
""".format(
    "".join([
        """
        <bundle id="scale{0}-bundle">
            <meta_attributes id="scale{0}-bundle-meta_attributes">
                <nvpair id="scale{0}-bundle-meta_attributes-target-role"
                    name="target-role" value="Stopped"
                />
            </meta_attributes>
            <docker run-command="/usr/sbin/pacemaker-remoted"
                image="user:remote"
                options="--user=root --log-driver=journald" replicas="20"
            />
            <network control-port="{0}"/>
            <storage>
                <storage-mapping target-dir="/dev/log"
                    id="dev-log-{0}" source-dir="/dev/log"
                />
            </storage>
        </bundle>
        """.format(i) for i in range(20000)
    ]),

    "".join([
        """
        <node_state id="{1}" uname="c09-h05-r630" in_ccm="true" crmd="online"
            crm-debug-origin="do_update_resource" join="member"
            expected="member"
        >
            <transient_attributes id="{1}">
                <instance_attributes id="status-{1}"/>
            </transient_attributes>
            <lrm id="1">
                <lrm_resources>{0}</lrm_resources>
            </lrm>
        </node_state>
        """.format("".join([
            """
            <lrm_resource id="scale13-bundle-{0}" type="remote" class="ocf"
                provider="pacemaker" container="scale13-bundle-docker-{0}"
            >
                <lrm_rsc_op id="scale13-bundle-{0}_last_0"
                    operation_key="scale13-bundle-{0}_start_0"
                    operation="start" crm-debug-origin="do_update_resource"
                    crm_feature_set="3.0.14"
                    transition-key="2957:15:0:2459ea96-7c1d-4276-9c21-828061199"
                    transition-magic="0:0;2957:15:0:2459ea96-7c1d-4276-9c21-828"
                    on_node="c09-h05-r630" call-id="2223" rc-code="0"
                    op-status="0" interval="0" last-run="1509318692"
                    last-rc-change="1509318692" exec-time="0" queue-time="0"
                    op-digest="802dc0edf5e736d13a41ac47626295eb"
                    op-force-restart=" reconnect_interval  port "
                    op-restart-digest="e38862dec2edf868edfcb2d64d77ff55"
                />
                <lrm_rsc_op id="scale13-bundle-{0}_monitor_60000"
                    operation_key="scale13-bundle-1{0}_monitor_60000"
                    operation="monitor" crm-debug-origin="do_update_resource"
                    crm_feature_set="3.0.14"
                    transition-key="2994:16:0:2459ea96-7c1d-4276-9c21-828061199"
                    transition-magic="0:0;2994:16:0:2459ea96-7c1d-4276-9c21-828"
                    on_node="c09-h05-r630" call-id="2264" rc-code="0"
                    op-status="0" interval="60000" last-rc-change="1509318915"
                    exec-time="0" queue-time="0"
                    op-digest="3b2ba04195253e454b50aa4a340af042"
                />
            </lrm_resource>
            """.format("{0}-{1}".format(i, j))
            for j in range(98)
        ]), i)
        for i in range(5)
    ])
)



from unittest import TestCase

from pcs.lib.corosync.node import (
    CorosyncNode,
    CorosyncNodeAddress,
)


class AddrsPlain(TestCase):
    @staticmethod
    def _fixture_no_addrs():
        return CorosyncNode("node1", [], "1")

    @staticmethod
    def _fixture_one_addr():
        return CorosyncNode(
            "node1",
            [CorosyncNodeAddress("10.0.0.1", "1")],
            "1"
        )

    @staticmethod
    def _fixture_several_addrs():
        return CorosyncNode(
            "node1",
            [
                CorosyncNodeAddress("10.0.0.0", "0"),
                CorosyncNodeAddress("10.0.0.1", "1"),
                CorosyncNodeAddress("10.0.0.4", "4"),
                CorosyncNodeAddress("10.0.0.3", "3"),
            ],
            "1"
        )

    def test_no_addrs_set_get_one(self):
        node = self._fixture_no_addrs()
        self.assertEqual(None, node.addr_plain_for_link("0"))

    def test_no_addrs_set_get_all(self):
        node = self._fixture_no_addrs()
        self.assertEqual([], node.addrs_plain())

    def test_no_addrs_set_get_all_except(self):
        node = self._fixture_no_addrs()
        self.assertEqual([], node.addrs_plain(except_link="0"))

    def test_one_addr_set_get_it(self):
        node = self._fixture_one_addr()
        self.assertEqual("10.0.0.1", node.addr_plain_for_link("1"))

    def test_one_addr_set_get_another(self):
        node = self._fixture_one_addr()
        self.assertEqual(None, node.addr_plain_for_link(1))
        self.assertEqual(None, node.addr_plain_for_link("2"))

    def test_one_addr_set_get_all(self):
        node = self._fixture_one_addr()
        self.assertEqual(["10.0.0.1"], node.addrs_plain())

    def test_one_addr_set_get_all_except_it(self):
        node = self._fixture_one_addr()
        self.assertEqual([], node.addrs_plain(except_link="1"))

    def test_one_addr_set_get_all_except_another(self):
        node = self._fixture_one_addr()
        self.assertEqual(["10.0.0.1"], node.addrs_plain(except_link=1))
        self.assertEqual(["10.0.0.1"], node.addrs_plain(except_link="2"))

    def test_several_addrs_set_get_one_of_them(self):
        node = self._fixture_several_addrs()
        self.assertEqual("10.0.0.1", node.addr_plain_for_link("1"))

    def test_several_addrs_set_get_another(self):
        node = self._fixture_several_addrs()
        self.assertEqual(None, node.addr_plain_for_link(1))
        self.assertEqual(None, node.addr_plain_for_link("2"))

    def test_several_addrs_set_get_all(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain()
        )

    def test_several_addrs_set_get_all_except_one(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link="1")
        )

    def test_several_addrs_set_get_all_except_another(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link=1)
        )
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link="2")
        )



from lxml import etree

from pcs.common import report_codes
from pcs.lib import reports
from pcs.lib.cib import resource
from pcs.lib.cib.constraint import resource_set
from pcs.lib.cib.tools import (
    find_unique_id,
    find_element_by_tag_and_id,
)
from pcs.lib.errors import LibraryError, ReportItemSeverity
from pcs.lib.xml_tools import (
    export_attributes,
    find_parent,
)


def _validate_attrib_names(attrib_names, options):
    invalid_names = [
        name for name in options.keys()
        if name not in attrib_names
    ]
    if invalid_names:
        raise LibraryError(
            reports.invalid_options(invalid_names, attrib_names, None)
        )

def find_valid_resource_id(report_processor, cib, in_clone_allowed, _id):
    parent_tags = resource.clone.ALL_TAGS + [resource.bundle.TAG]
    resource_element = find_element_by_tag_and_id(
        parent_tags + [resource.primitive.TAG, resource.group.TAG],
        cib,
        _id,
    )

    if resource_element.tag in parent_tags:
        return resource_element.attrib["id"]

    clone = find_parent(resource_element, parent_tags)
    if clone is None:
        return resource_element.attrib["id"]

    if in_clone_allowed:
        report_processor.process(
            reports.resource_for_constraint_is_multiinstance(
                resource_element.attrib["id"],
                "clone" if clone.tag == "master" else clone.tag,
                clone.attrib["id"],
                ReportItemSeverity.WARNING,
            )
        )
        return resource_element.attrib["id"]

    raise LibraryError(reports.resource_for_constraint_is_multiinstance(
        resource_element.attrib["id"],
        "clone" if clone.tag == "master" else clone.tag,
        clone.attrib["id"],
        ReportItemSeverity.ERROR,
        #repair to clone is workaround for web ui, so we put only information
        #about one forceable possibility
        forceable=report_codes.FORCE_CONSTRAINT_MULTIINSTANCE_RESOURCE
    ))

def prepare_options(attrib_names, options, create_id_fn, validate_id):
    _validate_attrib_names(attrib_names+("id",), options)
    options = options.copy()

    if "id" not in options:
        options["id"] = create_id_fn()
    else:
        validate_id(options["id"])
    return options

def export_with_set(element):
    return {
        "resource_sets": [
            resource_set.export(resource_set_item)
            for resource_set_item in element.findall(".//resource_set")
        ],
        "options": export_attributes(element),
    }

def export_plain(element):
    return {"options": export_attributes(element)}

def create_id(cib, type_prefix, resource_set_list):
    _id = "pcs_" +type_prefix +"".join([
        "_set_"+"_".join(id_set)
        for id_set in resource_set.extract_id_set_list(resource_set_list)
    ])
    return find_unique_id(cib, _id)

def have_duplicate_resource_sets(element, other_element):
    get_id_set_list = lambda element: [
        resource_set.get_resource_id_set_list(resource_set_item)
        for resource_set_item in element.findall(".//resource_set")
    ]
    return get_id_set_list(element) == get_id_set_list(other_element)

def check_is_without_duplication(
    report_processor,
    constraint_section, element, are_duplicate, export_element,
    duplication_alowed=False
):
    duplicate_element_list = [
        duplicate_element
        for duplicate_element in constraint_section.findall(".//"+element.tag)
        if(
            element is not duplicate_element
            and
            are_duplicate(element, duplicate_element)
        )
    ]
    if not duplicate_element_list:
        return

    report_processor.process(reports.duplicate_constraints_exist(
        element.tag,
        [
            export_element(duplicate_element)
            for duplicate_element in duplicate_element_list
        ],
        ReportItemSeverity.WARNING if duplication_alowed
            else ReportItemSeverity.ERROR,
        forceable=None if duplication_alowed
            else report_codes.FORCE_CONSTRAINT_DUPLICATE,
    ))

def create_with_set(constraint_section, tag_name, options, resource_set_list):
    if not resource_set_list:
        raise LibraryError(reports.empty_resource_set_list())
    element = etree.SubElement(constraint_section, tag_name)
    element.attrib.update(options)
    for resource_set_item in resource_set_list:
        resource_set.create(element, resource_set_item)
    return element



from pcs.lib import resource_agent
from pcs.lib.commands.resource_agent import _complete_agent_list


def list_agents(lib_env, describe=True, search=None):
    """
    List all stonith agents on the local host, optionally filtered and described
    bool describe load and return agents' description as well
    string search return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()
    agent_names = resource_agent.list_stonith_agents(runner)
    return _complete_agent_list(
        runner,
        agent_names,
        describe,
        search,
        resource_agent.StonithAgent
    )


def describe_agent(lib_env, agent_name):
    """
    Get agent's description (metadata) in a structure
    string agent_name name of the agent (not containing "stonith:" prefix)
    """
    agent = resource_agent.find_valid_stonith_agent_by_name(
        lib_env.report_processor,
        lib_env.cmd_runner(),
        agent_name,
        absent_agent_supported=False
    )
    return agent.get_full_info()



from unittest import TestCase

from pcs_test.tools.assertions import (
    assert_raise_library_error,
)
from pcs_test.tools.misc import get_test_resource as rc

from pcs.common import report_codes
from pcs.lib.errors import ReportItemSeverity as severity

import pcs.lib.corosync.config_facade as lib
from pcs.lib.corosync import constants

class FromStringTest(TestCase):
    def test_success(self):
        with open(rc("corosync.conf")) as a_file:
            config = a_file.read()
        facade = lib.ConfigFacade.from_string(config)
        self.assertEqual(facade.__class__, lib.ConfigFacade)
        self.assertEqual(facade.config.export(), config)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_parse_error_missing_brace(self):
        # pylint: disable=no-self-use
        config = "section {"
        assert_raise_library_error(
            lambda: lib.ConfigFacade.from_string(config),
            (
                severity.ERROR,
                report_codes.PARSE_ERROR_COROSYNC_CONF_MISSING_CLOSING_BRACE,
                {}
            )
        )

    def test_parse_error_unexpected_brace(self):
        # pylint: disable=no-self-use
        config = "}"
        assert_raise_library_error(
            lambda: lib.ConfigFacade.from_string(config),
            (
                severity.ERROR,
                report_codes.PARSE_ERROR_COROSYNC_CONF_UNEXPECTED_CLOSING_BRACE,
                {}
            )
        )


class GetSimpleValueMixin():
    def assert_value(self, value, config):
        facade = lib.ConfigFacade.from_string(config)
        self.assertEqual(value, self.getter(facade))
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class GetClusterNameTest(GetSimpleValueMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_cluster_name()

    def test_no_name(self):
        self.assert_value(
            "",
            ""
        )

    def test_empty_name(self):
        self.assert_value(
            "",
            "totem {\n cluster_name:\n}\n"
        )

    def test_one_name(self):
        self.assert_value(
            "test",
            "totem {\n cluster_name: test\n}\n"
        )

    def test_more_names(self):
        self.assert_value(
            "TEST",
            "totem {\n cluster_name: test\n cluster_name: TEST\n}\n"
        )

    def test_more_sections(self):
        self.assert_value(
            "TEST",
            "totem{\n cluster_name: test\n}\ntotem{\n cluster_name: TEST\n}\n"
        )


class GetTransport(GetSimpleValueMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_transport()

    def test_no_name(self):
        self.assert_value(
            constants.TRANSPORT_DEFAULT,
            ""
        )

    def test_empty_name(self):
        self.assert_value(
            constants.TRANSPORT_DEFAULT,
            "totem {\n transport:\n}\n"
        )

    def test_one_name(self):
        self.assert_value(
            "udp",
            "totem {\n transport: udp\n}\n"
        )

    def test_more_names(self):
        self.assert_value(
            "udpu",
            "totem {\n transport: udp\n transport: udpu\n}\n"
        )

    def test_more_sections(self):
        self.assert_value(
            "udpu",
            "totem{\n transport: udp\n}\ntotem{\n transport: udpu\n}\n"
        )


class GetIpVersion(GetSimpleValueMixin, TestCase):
    transport_ip_list = [
        ("udp", constants.IP_VERSION_4),
        ("udpu", constants.IP_VERSION_64),
        ("knet", constants.IP_VERSION_64),
    ]

    @staticmethod
    def getter(facade):
        return facade.get_ip_version()

    def test_empty_config(self):
        self.assert_value(
            constants.IP_VERSION_64,
            ""
        )

    def test_no_name(self):
        for transport, ip in self.transport_ip_list:
            with self.subTest(transport=transport, ip=ip):
                self.assert_value(
                    ip,
                    f"totem {{\n transport: {transport}\n}}\n"
                )

    def test_no_value(self):
        for transport, ip in self.transport_ip_list:
            with self.subTest(transport=transport, ip=ip):
                self.assert_value(
                    ip,
                    f"totem {{\n transport: {transport}\n ip_version:\n}}\n"
                )

    def test_one_name(self):
        self.assert_value(
            "ipv4-6",
            "totem {\n transport: udp\n ip_version: ipv4-6\n}\n"
        )

    def test_more_names(self):
        self.assert_value(
            "ipv6",
            "totem {\ntransport: udp\nip_version: ipv4-6\nip_version: ipv6\n}\n"
        )

    def test_more_sections(self):
        self.assert_value(
            "ipv6",
            "totem {\n transport: knet\n ip_version: ipv4-6\n}\n"
                "totem {\n ip_version: ipv6\n}\n"
        )



from pcs import (
    prop,
    usage,
)
from pcs.cli.common.routing import create_router


property_cmd = create_router(
    {
        "help": lambda _lib, _argv, _modifiers: usage.property(_argv),
        "set": prop.set_property,
        "unset": prop.unset_property,
        "list": prop.list_property,
        "show": prop.list_property,
        "get_cluster_properties_definition":
            prop.print_cluster_properties_definition,
    },
    ["property"],
    default_cmd="list"
)



import json
import logging
from base64 import b64encode
from unittest import TestCase, mock
from urllib.parse import urlencode

from tornado.httputil import HTTPServerRequest
from tornado.testing import AsyncTestCase, gen_test
from tornado.web import HTTPError

from pcs_test.tools.misc import create_patcher, get_test_resource as rc

from pcs.daemon import ruby_pcsd

# Don't write errors to test output.
logging.getLogger("pcs.daemon").setLevel(logging.CRITICAL)

def create_wrapper():
    return ruby_pcsd.Wrapper(
        rc("/path/to/gem_home"),
        rc("/path/to/pcsd/cmdline/entry"),
    )

def create_http_request():
    return HTTPServerRequest(
        method="POST",
        uri="/pcsd/uri",
        headers={"Cookie": "cookie1=first;cookie2=second"},
        body=str.encode(urlencode({"post-key": "post-value"})),
        host="pcsd-host:2224"
    )

class GetSinatraRequest(TestCase):
    def test_translate_request(self):
        # pylint: disable=invalid-name
        self.maxDiff = None
        self.assertEqual(
            create_wrapper().get_sinatra_request(create_http_request()),
            {
                'env': {
                    'HTTPS': 'off',
                    'HTTP_ACCEPT': '*/*',
                    'HTTP_COOKIE': 'cookie1=first;cookie2=second',
                    'HTTP_HOST': 'pcsd-host:2224',
                    'HTTP_VERSION': 'HTTP/1.0',
                    'PATH_INFO': '/pcsd/uri',
                    'QUERY_STRING': '',
                    'REMOTE_ADDR': None, # It requires complicated request args
                    'REMOTE_HOST': 'pcsd-host:2224',
                    'REQUEST_METHOD': 'POST',
                    'REQUEST_PATH': '/pcsd/uri',
                    'REQUEST_URI': 'http://pcsd-host:2224/pcsd/uri',
                    'SCRIPT_NAME': '',
                    'SERVER_NAME': 'pcsd-host',
                    'SERVER_PORT': 2224,
                    'SERVER_PROTOCOL': 'HTTP/1.0',
                    'rack.input': 'post-key=post-value'
                }
            }
        )

patch_ruby_pcsd = create_patcher(ruby_pcsd)

class RunRuby(AsyncTestCase):
    def setUp(self):
        self.stdout = ""
        self.stderr = ""
        self.exit_status = 0
        self.request = self.create_request()
        self.wrapper = create_wrapper()
        patcher = mock.patch.object(
            self.wrapper,
            "send_to_ruby",
            self.send_to_ruby
        )
        self.addCleanup(patcher.stop)
        patcher.start()
        super().setUp()

    async def send_to_ruby(self, request_json):
        self.assertEqual(json.loads(request_json), self.request)
        return self.stdout, self.stderr, self.exit_status

    @staticmethod
    def create_request(_type=ruby_pcsd.SYNC_CONFIGS):
        return {"type": _type}

    def set_run_result(self, run_result):
        self.stdout = json.dumps({**run_result, "logs": []})

    def assert_sinatra_result(self, result, headers, status, body):
        self.assertEqual(result.headers, headers)
        self.assertEqual(result.status, status)
        self.assertEqual(result.body, str.encode(body))

    @gen_test
    def test_correct_sending(self):
        run_result = {"next": 10}
        self.set_run_result(run_result)
        result = yield self.wrapper.run_ruby(ruby_pcsd.SYNC_CONFIGS)
        self.assertEqual(result["next"], run_result["next"])

    @gen_test
    def test_error_from_ruby(self):
        with self.assertRaises(HTTPError):
            yield self.wrapper.run_ruby(ruby_pcsd.SYNC_CONFIGS)

    @gen_test
    def test_sync_config_shortcut_success(self):
        _next = 10
        self.set_run_result({"next": _next})
        result = yield self.wrapper.sync_configs()
        self.assertEqual(result, _next)

    @patch_ruby_pcsd("now", return_value=0)
    @gen_test
    def test_sync_config_shorcut_fail(self, now):
        # pylint: disable=unused-argument
        result = yield self.wrapper.sync_configs()
        self.assertEqual(result, ruby_pcsd.DEFAULT_SYNC_CONFIG_DELAY)

    @gen_test
    def test_request_remote(self):
        headers = {"some": "header"}
        status = 200
        body = "content"
        self.set_run_result({
            "headers": headers,
            "status": status,
            "body": b64encode(str.encode(body)).decode(),
        })
        http_request = create_http_request()
        self.request = {
            **self.create_request(ruby_pcsd.SINATRA_REMOTE),
            **self.wrapper.get_sinatra_request(http_request),
        }
        result = yield self.wrapper.request_remote(http_request)
        self.assert_sinatra_result(result, headers, status, body)

    @gen_test
    def test_request_gui(self):
        headers = {"some": "header"}
        status = 200
        body = "content"

        user = "user"
        groups = ["hacluster"]
        is_authenticated = True

        self.set_run_result({
            "headers": headers,
            "status": status,
            "body": b64encode(str.encode(body)).decode(),
        })
        http_request = create_http_request()
        self.request = {
            **self.create_request(ruby_pcsd.SINATRA_GUI),
            **self.wrapper.get_sinatra_request(http_request),
            "session": {
                "username": user,
                "groups": groups,
                "is_authenticated": is_authenticated,
            }
        }
        result = yield self.wrapper.request_gui(
            http_request,
            user=user,
            groups=groups,
            is_authenticated=is_authenticated,
        )
        self.assert_sinatra_result(result, headers, status, body)

class ProcessResponseLog(TestCase):
    @patch_ruby_pcsd("log.from_external_source")
    @patch_ruby_pcsd("next", mock.Mock(return_value=1))
    def test_put_correct_values_to_log(self, from_external_source):
        # pylint: disable=no-self-use
        ruby_pcsd.process_response_logs([{
            "level": "FATAL",
            "timestamp_usec": 1234567890,
            "message": "ruby_message",
        }])
        from_external_source.assert_called_once_with(
            level=logging.CRITICAL,
            created=1234.56789,
            usecs=567890,
            message="ruby_message",
            group_id=1,
        )



from pcs.common import report_codes
from pcs.common.node_communicator import Request
from pcs.lib import reports
from pcs.lib.node_communication import response_to_report_item
from pcs.lib.errors import (
    LibraryError,
    ReportItemSeverity,
)


class CommunicationCommandInterface:
    """
    Interface for all communication commands.
    """
    def get_initial_request_list(self):
        """
        Returns an initial list of Request object.
        """
        raise NotImplementedError()

    def on_response(self, response):
        """
        Process received response. Returns list of new Request that should be
        added to the executing queue.

        Response response -- a response to be processed
        """
        raise NotImplementedError()

    def on_complete(self):
        """
        Runs after all reqests finished.
        """
        raise NotImplementedError()

    def before(self):
        """
        Runs before executing the requests.
        """
        raise NotImplementedError()

    @property
    def error_list(self):
        """
        List of errors which occured during running the requests.
        """
        raise NotImplementedError()


def run(communicator, cmd):
    """
    Run communication command. Returns return value of method on_complete() of
    communcation command after run.

    NodeCommunicator communicator -- object used for communication
    CommunicationCommandInterface cmd
    """
    cmd.before()
    communicator.add_requests(cmd.get_initial_request_list())
    for response in communicator.start_loop():
        extra_requests = cmd.on_response(response)
        if extra_requests:
            communicator.add_requests(extra_requests)
    return cmd.on_complete()


def run_and_raise(communicator, cmd):
    """
    Run communication command. Returns return value of method on_complete() of
    communcation command after run.
    Raises LibraryError (with no report item) when some errors occured while
    running communication command.

    NodeCommunicator communicator -- object used for communication
    CommunicationCommandInterface cmd
    """
    to_return = run(communicator, cmd)
    if cmd.error_list:
        raise LibraryError()
    return to_return


class RunRemotelyBase(CommunicationCommandInterface):
    """
    Abstract class for communication commands. This class provides methods for
    reporting.
    """
    # pylint: disable=abstract-method
    _report_pcsd_too_old_on_404 = False

    def __init__(self, report_processor):
        self.__report_processor = report_processor
        self._error_list = []

    def _get_response_report(self, response):
        """
        Convert specified response to report item. Returns None if the response
        has no failures.

        Response response -- a response to be converted
        """
        return response_to_report_item(
            response,
            report_pcsd_too_old_on_404=self._report_pcsd_too_old_on_404,
        )

    def _report_list(self, report_list):
        """
        Send reports from report_list to the report processor.

        list report_list -- list of ReportItem objects
        """
        self._error_list.extend(
            self.__report_processor.report_list(report_list)
        )

    def _report(self, report):
        """
        Send specified report to the report processor.

        ReportItem report -- report which will be reported
        """
        self._report_list([report])

    def _process_response(self, response):
        """
        Process received response. Returns list of new Request that should be
        added to the executing queue. If no new Request should be added, there
        is no need to return empty list.

        Response response -- a response to be processed
        """
        raise NotImplementedError()

    def on_response(self, response):
        returned = self._process_response(response)
        return returned if returned else []

    def on_complete(self):
        return None

    def before(self):
        pass

    @property
    def error_list(self):
        return self._error_list


class StrategyBase:
    """
    Abstract base class of the communication strategies. Always use at most one
    strategy mixin in the communication commands classes.
    """
    def _prepare_initial_requests(self):
        """
        Returns list of all Request objects which should be run. Full
        implementation of strategy will use this list for creating initial
        request list and others.
        """
        raise NotImplementedError()

    def get_initial_request_list(self):
        """
        This method has to be implemented by the descendants.
        """
        raise NotImplementedError()


class OneByOneStrategyMixin(StrategyBase):
    """
    Communication strategy in which requests are executed one by one. So only
    one request from _prepare_initial_requests is chosen as initial request
    list. Other requests are then available by calling method _get_next_list.
    """
    #pylint: disable=abstract-method
    __iter = None
    __successful = False

    def get_initial_request_list(self):
        """
        Returns only first request from _prepare_initial_requests.
        """
        self.__iter = iter(self._prepare_initial_requests())
        return self._get_next_list()

    def _get_next_list(self):
        """
        Returns a list which contains another Request object from
        _prepare_initial_requests. Raises StopIteration when there is no other
        request left.
        """
        try:
            return [next(self.__iter)]
        except StopIteration:
            return []


class AllAtOnceStrategyMixin(StrategyBase):
    """
    Communication strategy in which all requests are executed at once in
    parallel.
    """
    #pylint: disable=abstract-method
    def get_initial_request_list(self):
        return self._prepare_initial_requests()


class MarkSuccessfulMixin:
    __successful = False

    def _on_success(self):
        self.__successful = True

    def on_complete(self):
        if not self.__successful:
            self._report(reports.unable_to_perform_operation_on_any_node())


class AllSameDataMixin:
    """
    Communication command mixin which adds common methods for commands where
    requests to all targets have the same data.
    """
    __targets = None

    def _get_request_data(self):
        """
        Returns RequestData object to use as data for requests to all targets.
        """
        raise NotImplementedError()

    def _prepare_initial_requests(self):
        return [
            Request(target, self._get_request_data())
            for target in self.__target_list
        ]

    def add_request(self, target):
        """
        Add target to which request will be send.

        RequestTarget target -- target that will be added.
        """
        self.set_targets([target])

    def set_targets(self, target_list):
        """
        Add targets to which requests will be send.

        list target_list -- RequestTarget list
        """
        self.__target_list.extend(target_list)

    @property
    def __target_list(self):
        if self.__targets is None:
            self.__targets = []
        return self.__targets

    @property
    def _target_list(self):
        """
        List of RequestTarget to which request will be send.
        """
        return list(self.__target_list)

    @property
    def _target_label_list(self):
        return [target.label for target in self.__target_list]



class SimpleResponseProcessingMixin:
    """
    Communication command mixin which adds common response processing. When
    request fails error/warning will be reported. Otherwise _get_success_report
    will be reported.
    """
    def _get_success_report(self, node_label):
        """
        Returns ReportItem which should be reported when request was
        successfull.

        string node_label -- node identifier on which request was successful
        """
        raise NotImplementedError()

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is None:
            report = self._get_success_report(response.request.target.label)
        self._report(report)


class SimpleResponseProcessingNoResponseOnSuccessMixin:
    """
    Communication command mixin which adds common response processing. When
    request fails error/warning will be reported.
    """
    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)


class SkipOfflineMixin:
    """
    Communication command mixin which simplifies handling of forcing skip
    offline nodes. This mixin provides method _set_skip_offline which should be
    called from __init__ of the descendants. Then report item from response
    returned from _get_response_report is set accordingly to value of
    skip_offline_targets.
    """
    _failure_severity = ReportItemSeverity.ERROR
    _failure_forceable = None
    _report_pcsd_too_old_on_404: bool

    def _set_skip_offline(
        self, skip_offline_targets, force_code=report_codes.SKIP_OFFLINE_NODES
    ):
        """
        Set value of skip_offline_targets flag.

        boolean skip_offline_targets
        """
        self._failure_forceable = force_code
        if skip_offline_targets:
            self._failure_severity = ReportItemSeverity.WARNING
            self._failure_forceable = None

    def _get_response_report(self, response):
        return response_to_report_item(
            response,
            severity=self._failure_severity,
            forceable=self._failure_forceable,
            report_pcsd_too_old_on_404=self._report_pcsd_too_old_on_404,
        )



from unittest import TestCase

from pcs.cli.common import printable_tree as lib


class Node:
    def __init__(self, title, detail, is_leaf, members):
        self.title = title
        self.detail = detail
        self.is_leaf = is_leaf
        self.members = members

    def get_title(self, verbose):
        detail = " (verbose)" if verbose else ""
        return f"{self.title}{detail}"

def node(an_id, detail=0, leaf=False, members=None):
    return Node(
        f"{an_id}-title",
        [f"{an_id}-detail{i}" for i in range(detail)],
        leaf,
        members=members or [],
    )

class TreeToLines(TestCase):
    def test_verbose(self):
        self.assertEqual(
            ["l0-title (verbose)"], lib.tree_to_lines(node("l0"), verbose=True)
        )

    def test_empty(self):
        self.assertEqual(["l0-title"], lib.tree_to_lines(node("l0")))

    def test_empty_leaf(self):
        self.assertEqual(
            ["l0-title [displayed elsewhere]"],
            lib.tree_to_lines(node("l0", leaf=True)),
        )

    def test_detail_simple(self):
        self.assertEqual(
            [
                "l0-title",
                "   l0-detail0",
            ],
            lib.tree_to_lines(node("l0", 1)),
        )

    def test_detail(self):
        self.assertEqual(
            [
                "l0-title",
                "   l0-detail0",
                "   l0-detail1",
                "   l0-detail2",
            ],
            lib.tree_to_lines(node("l0", 3)),
        )

    def test_detail_leaf(self):
        self.assertEqual(
            ["l0-title [displayed elsewhere]"],
            lib.tree_to_lines(node("l0", 3, True)),
        )

    def test_one_member(self):
        self.assertEqual(
            [
                "l0-title",
                "`- l1-title",
            ],
            lib.tree_to_lines(node("l0", members=[node("l1")])),
        )

    def test_one_member_leaf(self):
        self.assertEqual(
            ["l0-title [displayed elsewhere]"],
            lib.tree_to_lines(node("l0", leaf=True, members=[node("l1")])),
        )

    def test_multiple_members(self):
        self.assertEqual(
            [
                "l0-title",
                "|- l1-title",
                "|- l2-title",
                "`- l3-title",
            ],
            lib.tree_to_lines(
                node("l0", members=[node("l1"), node("l2"), node("l3")])
            ),
        )

    def test_multiple_members_detail(self):
        self.assertEqual(
            [
                "l0-title",
                "|  l0-detail0",
                "|  l0-detail1",
                "|- l1-title",
                "|- l2-title",
                "`- l3-title",
            ],
            lib.tree_to_lines(
                node(
                    "l0", detail=2, members=[node("l1"), node("l2"), node("l3")]
                )
            ),
        )

    def test_multiple_members_detail_leaf(self):
        self.assertEqual(
            ["l0-title [displayed elsewhere]"],
            lib.tree_to_lines(
                node("l0", 2, True, [node("l1"), node("l2"), node("l3")])
            ),
        )

    def test_complex_tree_wide(self):
        self.assertEqual(
            [
                "0-title",
                "|  0-detail0",
                "|  0-detail1",
                "|- 00-title",
                "|     00-detail0",
                "|     00-detail1",
                "|- 01-title",
                "|  |- 010-title",
                "|  |  `- 0100-title [displayed elsewhere]",
                "|  `- 011-title",
                "|- 02-title",
                "|  |  02-detail0",
                "|  |  02-detail1",
                "|  |  02-detail2",
                "|  |- 020-title",
                "|  `- 021-title",
                "|     |  021-detail0",
                "|     |- 0210-title [displayed elsewhere]",
                "|     |- 0211-title",
                "|     |     0211-detail0",
                "|     |     0211-detail1",
                "|     `- 0212-title",
                "|        `- 02120-title",
                "|- 03-title",
                "|  `- 030-title",
                "|     `- 0300-title",
                "|- 04-title [displayed elsewhere]",
                "`- 05-title",
                "      05-detail0",
            ],
            lib.tree_to_lines(
                node(
                    "0",
                    2,
                    members=[
                        node("00", 2),
                        node(
                            "01", members=[
                                node("010", members=[node("0100", leaf=True)]),
                                node("011"),
                            ],
                        ),
                        node(
                            "02", 3, members=[
                                node("020"),
                                node(
                                    "021", 1, members=[
                                        node("0210", leaf=True),
                                        node("0211", 2),
                                        node("0212", members=[node("02120")]),
                                    ],
                                ),
                            ],
                        ),
                        node(
                            "03", members=[node("030", members=[node("0300")])],
                        ),
                        node("04", leaf=True),
                        node("05", 1),
                    ],
                )
            ),
        )

    def test_complex_tree_deep(self):
        self.assertEqual(
            [
                "0-title",
                "|  0-detail0",
                "`- 00-title",
                "   |- 000-title",
                "   |     000-detail0",
                "   |     000-detail1",
                "   `- 001-title",
                "      |- 0010-title [displayed elsewhere]",
                "      |- 0011-title",
                "      |     0011-detail0",
                "      |     0011-detail1",
                "      `- 0012-title",
                "         `- 00120-title",
            ],
            lib.tree_to_lines(
                node("0", 1, members=[
                    node("00", members=[
                        node("000", 2),
                        node("001", members=[
                            node("0010", leaf=True),
                            node("0011", 2),
                            node("0012", members=[node("00120")]),
                        ]),
                    ]),
                ])
            ),
        )
